------------------------------------------------------------------------------
--  Ada95 Interface to Oracle RDBMS                                         --
--  Copyright (C) 2000-2001 Dmitriy Anisimkov.                              --
--  License text and authors contact information could be found in oci.ads  --
------------------------------------------------------------------------------
with Text_IO,System.Address_Image; use Text_IO,System;

package body OCI.Environments is

    use type sword,ub4;
    use type OCIHandle;
   
  
   protected body Counter is

      procedure Add (It : Integer) is
      begin
         Val := Val + It;
      end;
      
      function Value return Integer is
      begin
        return Val;
      end;
      
   end Counter;
   
   function malloc (Size : Integer) return System.Address;
   pragma Import (C, malloc, "__gnat_malloc");

   procedure free (Address : System.Address);
   pragma Import (C, free, "__gnat_free");

   function Realloc (Ptr : Address; Size : Integer) return System.Address;
   pragma Import (C, Realloc, "__gnat_realloc");

   function Alloc_Memory (ctxp : System.Address;
          Size : Integer) return System.Address;

   function Realloc_Memory (ctxp : System.Address; Memptr : System.Address;
          Size : Integer) return System.Address;

   procedure Free_Memory (ctxp : System.Address; Memptr : System.Address);
   
   pragma Convention (C, Alloc_Memory);
   pragma Convention (C, Realloc_Memory);
   pragma Convention (C, Free_Memory);
   
   function Alloc_Memory (ctxp : System.Address;
          Size : Integer) return System.Address
   is
      use System;
   begin
      Counter.Add (1);
      Last_Size := Size;
      return Malloc (Size);
   end;

   function Realloc_Memory (ctxp : System.Address; Memptr : System.Address;
          Size : Integer) return System.Address
   is
   begin
      if Size = 0 then
         Counter.Add (-1);
      end if;
--      Text_IO.Put_Line ("realloc " & Integer'Image (Size));
      return Realloc (Memptr, Size);
   end;

   procedure Free_Memory (ctxp : System.Address; Memptr : System.Address) is
     use type System.Address;
   begin
      Counter.Add (-1);
--      Text_IO.Put ("Free memory");
      Free (Memptr);
--      Text_IO.Put_Line (" done.");
   end;   

  ---------------------------------------------------------------------
          
  function Create return Thread_Environment is
      Env : aliased OCIEnv := OCIEnv(Empty_Handle);
      Rc : Sword;
      Environment : Thread_Environment;
  begin
--     Put_Line("Environment create");
     Rc := OCIEnvCreate (Env'access,
          OCI_THREADED + OCI_OBJECT
        ,Malocfp  => Alloc_Memory'Address
        ,Ralocfp  => Realloc_Memory'Address
        ,Mfreefp  => Free_Memory'Address
      );
--     Put_Line("Environment created");
     if Rc /= OCI_SUCCESS then
--        Put_Line("Environment creation error");
        raise ENVIRONMENT_CREATION_ERROR;
     end if;
     Environment.Handle := Env;
     return Environment;
  end;

  procedure Destroy   (Object : in out Thread_Environment) is
     RC : Sword;
  begin
     if Object.Handle /= OCIEnv (Empty_Handle) then
        RC := OCIHandleFree (OCIHandle (Object.Handle), OCI_HTYPE_ENV);
        if Rc /= OCI_SUCCESS then
          Put_Line ("return code on free environment " & SWord'Image (Rc));
        end if;
     end if;  
  end;

end OCI.Environments;

