------------------------------------------------------------------------------
--  Ada95 Interface to Oracle RDBMS                                         --
--  Copyright (C) 2000-2001 Dmitriy Anisimkov.                              --
--  License text and authors contact information could be found in oci.ads  --
------------------------------------------------------------------------------
with Ada.Streams;
with Ada.Strings.Unbounded;
with OCI.Thick.Connections;
with System.Address_To_Access_Conversions;

package OCI.Thick.Lobs is

   use Ada.Streams;
   use Connections;
   
   type Lob_Type is (Bin,File,Char
--,NChar unsupported
   );
   
   type Lob is new Handle_Reference with private;
   
   function Create_Temporary (Connect : Connection; ltype : Lob_Type := Char) return Lob;

   function Create_Empty (Connect : Connection; ltype : Lob_Type := Char) return Lob;

   procedure Create_Temporary (Loc : in out Lob; Cache : in Boolean := True);

   procedure Free_Temporary (Loc : in out Lob);
   
   type Count  is new Natural;
   subtype Positive_Count is Count range 1 .. Count'Last;		     
   
   subtype raw is Ada.Streams.Stream_Element_Array;

--   procedure Append(
--     dest : IN OUT LOB;
--     src : IN     LOB); 

--   procedure CLOSE (
--      loc :   IN OUT LOB); 

   procedure READ (
     loc    :  IN LOB;
     amount :  OUT Count;
     offset :  IN Positive_Count;
     buffer :  OUT RAW);

   procedure READ (
     loc    :  IN LOB;
     amount :  OUT Count;
     offset :  IN Positive_Count;
     buffer :  OUT String);

   procedure WRITE (
     loc : IN OUT LOB;
     offset : IN Positive_Count;
     buffer : IN RAW);

   procedure WRITE (
     loc : IN OUT LOB;
     offset : IN Positive_Count;
     buffer : IN String);

   function Length (Loc : in Lob) return Count;
   function To_String (Loc : in Lob) return String;
   function To_Unbounded_String (Loc : in Lob) return Ada.Strings.Unbounded.Unbounded_String;
   
   function Is_Init (Value : in Lob) return boolean;
   function Is_Open (Value : in Lob) return boolean;
       
   type Lob_Stream (Ltype : Lob_Type) is new Root_Stream_Type with private;

   procedure Create_Temporary (Stream : in out Lob_Stream; Connect : Connection);
   procedure Free_Temporary (Stream : in out Lob_Stream);
   function To_String (Loc : in Lob_Stream) return String;
   procedure Reset (Stream : in out Lob_Stream);
    
   procedure Read(Stream : in out Lob_Stream;
      Item   : out Stream_Element_Array;
	Last   : out Stream_Element_Offset);

   procedure Write(Stream : in out Lob_Stream;
      Item   : in Stream_Element_Array);
       
   function Get_Lob_Type (Loc : Lob) return Lob_Type;
   function Get_Bind_Address (Loc : Lob) return DVoid;

private

   use Reference_Finalization;
   
   pragma inline (To_String);
   
   package AC is new
        System.Address_To_Access_Conversions (OCIHandle);
        
   type Lob is new Handle_Reference with record
      Connect : Connection;
      Ltype : Lob_Type;
      locp : AC.Object_Pointer := null;
   end record;

   procedure Destroy (Object : in out Lob);
   
   type Lob_Stream (Ltype : Lob_Type) is new Root_Stream_Type with record
     loc : Lob;
     Position : Positive_Count := Positive_Count'First;
   end record;

end OCI.Thick.Lobs;