------------------------------------------------------------------------------
--  Ada95 Interface to Oracle RDBMS                                         --
--  Copyright (C) 2000-2001 Dmitriy Anisimkov.                              --
--  License text and authors contact information could be found in oci.ads  --
------------------------------------------------------------------------------
with OCI.Lib;
with OCI.Thread;
with Text_IO;
with Ada.Exceptions;
with System;

package body OCI.Thick.Strings is

   use Lib;

   use Interfaces.C;
   use Strings;
   use OCI.Thread;

   type Access_Size_T is access all size_t;

   type Buffer_Ptr is access all Char_Buffer;

   Type_Id : constant := OCI.Lib.SQLT_LNG;
   Indicator_Unchanged : constant := Sb2'First;

   String_Max_Size : constant := 32767;
   
   procedure Show (Str : String) is
   begin
--      Text_IO.Put_Line (Str);
      null;
   end Show;

   procedure Set_Last_Value (Value : Value_Ptr) is
   begin
      Thread.Attributes.Reference.Last_Value := Value;
   end;
   
   function Get_Last_Value return Value_Ptr is
   begin
      return Thread.Attributes.Reference.Last_Value;
   end;
   
   function Get_Buffer_Ptr return Buffer_Ptr is
   begin
      return Thread.Attributes.Reference.Buffer'Access;
   end;
  
   procedure Complete_Last_Value
   is
      Last_Value : Value_Ptr := Get_Last_Value;
      Buffer : Buffer_Ptr := Get_Buffer_Ptr;

      function String_Remain return String
      is
      begin
         return To_Ada (Buffer (0 .. Last_Value.alen - 1), Trim_Nul => False);
      end String_Remain;
  
   begin
      if Last_Value = null then
         return;
      end if;
      pragma Assert (not Last_Value.Complete);
--      if Last_Value.rcode /= OCI_SUCCESS then
--         Ada.Exceptions.Raise_Exception (LIB_ERROR'Identity, "rcode = " & Ub2'Image (Last_Value.rcode));
--      end if;
      pragma Debug (Show ("last value len " & size_t'Image (Last_Value.alen)));
      if Last_Value.Indicator /= Null_Indicator then
         if Last_Value.First_Out_Piece then
            if Last_Value.Indicator = Indicator_Unchanged then
               pragma Debug (Show ("Unchanged indicator"));
               Last_Value.Indicator := Last_Value.Prev_Indicator;
            else
               Last_Value.Value := To_Unbounded_String (String_Remain);
            end if;
         else
            pragma Debug (Show ("Not first last out"));
            Last_Value.Value := Last_Value.Value & String_Remain;
         end if;
       end if;

       Last_Value.First_Out_Piece := False;
       
       Last_Value.alen := 0;

       Last_Value.Complete := True;
       Set_Last_Value (null);
   end Complete_Last_Value;
   
   function Get_Value (Var : Var_Type'Class) return Unbounded_String is
   begin
      pragma Assert (Var.Complete);
      if OCI.Thick.Is_Null(Var) then
         raise NULL_VALUE;
      end if;
      return Var.Value;
   end;
   
--   procedure Finalize  (Object : in out Var_Type)
--   is
--   begin
--      if not Object.Complete then
--        pragma Assert (System."=" (Object'Address, Get_Last_Value.all'Address));
--        Set_Last_Value (null);
--        Object.Complete := true;
--      end if;
--      pragma Debug (Show ("final string"));
--      OCI.Thick.Finalize (Limited_Variable (Object));
--   end;
      
   function Get_Value (Var : Var_Type'Class) return String is
   begin
      return To_String (Get_Value (Var));
   end Get_Value;  

   procedure Set_Value (Var : in out Var_Type'Class; Value : Unbounded_String) is
   begin
      Var.Indicator := 0;
      Var.Value := Value;
      if not Var.Complete then
         Set_Last_Value (null);
         var.First_Out_Piece := False;
         Var.Complete := true;
      end if;
   end;
   
   procedure Set_Value (Var : in out Var_Type'Class; Value : String) is
   begin
      Set_Value (Var, To_Unbounded_String (Value));
   end;
      
   use type sb4;

   function Callback_In_Bind (
                             ictxp  : Value_Ptr;
                             bindp  : OCIBind;
                             iter   : Ub4;
                             index  : Ub4;
                             bufpp  : access Buffer_Ptr;
                             alenp  : access Ub4;
                             piecep : access Ub1;
                             indp   : access Sb2) return SWord;
                             
   function Callback_Out (Value : Value_Ptr;
                             bufpp   : access Buffer_ptr;
                             alenpp  : access Access_Size_T;
                             piecep  : access Ub1;
                             indpp   : access A_Sb2) return SWord;
                             
   function Callback_Out_Bind (
                             octxp   : Value_Ptr;
                             bindp   : OCIBind;
                             iter    : Ub4;
                             index   : Ub4;
                             bufpp   : access Buffer_ptr;
                             alenpp  : access Access_Size_T;
                             piecep  : access Ub1;
                             indpp   : access A_Sb2;
                             rcodepp : access A_Ub2) return SWord;
                                
   type A_A_Ub2 is access A_Ub2;                             

   function Callback_Define (
                                octxp   : Value_Ptr;
                                defnp   : OCIDefine;
                                iter    : Ub4;
                                bufpp   : access Buffer_Ptr;
                                alenpp  : access Access_Size_T;
                                piecep  : access Ub1;
                                indpp   : access A_Sb2;
                                rcodepp : A_A_Ub2) return SWord;

   pragma Convention (C, Callback_In_Bind);
   pragma Convention (C, Callback_Out_Bind);
   pragma Convention (C, Callback_Define);

   --  checks if remaining string, move it where need
   function Get_Buffer return Buffer_Ptr
   is
   begin
      Complete_Last_Value;
      return Get_Buffer_Ptr;
   end;
   
   function Callback_In_Bind (
                             ictxp  : Value_Ptr;
                             bindp  : OCIBind;
                             iter   : Ub4;
                             index  : Ub4;
                             bufpp  : access Buffer_ptr;
                             alenp  : access Ub4;
                             piecep : access Ub1;
                             indp   : access Sb2) return SWord
   is
      Value : Value_Ptr renames ictxp;
      Next_Position : Positive := Value.Position + Char_Buffer'Length;
      Next_Value_Position : Positive;
   begin
      pragma Debug (Show ("callback in " & Ub4'Image (alenp.all) & Sb2'Image (Value.Indicator)));

      indp.all := Value.Indicator;
        
      if Value.Value = Null_Unbounded_String 
      or Value.Indicator = Null_Indicator then
         alenp.all := 0;
         piecep.all := OCI_ONE_PIECE;
--         indp.all := Null_Indicator;
         pragma Debug ( Show ("null detected"));
      else

         if Next_Position <= Length (Value.Value) then
            Next_Value_Position := Next_Position;
            if Value.Position = 1 then
               piecep.all := OCI_FIRST_PIECE;
--               Text_IO.Put_Line ("OCI_FIRST_PIECE");
            else
               piecep.all := OCI_NEXT_PIECE;
--               Text_IO.Put_Line ("OCI_NEXT_PIECE");
            end if;
         else
            Next_Value_Position := 1;
            if Value.Position = 1 then
               piecep.all := OCI_ONE_PIECE;
--               Text_IO.Put_Line ("OCI_ONE_PIECE");
            else
               piecep.all := OCI_LAST_PIECE;
--               Text_IO.Put_Line ("OCI_LAST_PIECE");
            end if;
         end if;

         bufpp.all := Get_Buffer;

         declare
            Part : String := Slice (Value.Value,
                  Value.Position, Natural'Min 
                        (Next_Position - 1, Length(Value.Value)));
            Moved : Size_T;
         begin
            pragma debug (Show (Part));
            To_C (Part, bufpp.all.all, Moved, Append_Nul => False);
            alenp.all := Ub4 (Part'Length);
         end;
         
         Value.Position := Next_Value_Position;
         
--         Text_IO.Put_Line ("done in " & Ub4'Image (alenp.all));
         
      end if;
      pragma Debug ( Show ("end in callback"));
      return OCI_CONTINUE;
   exception
      when e: others =>
        Text_IO.Put_Line (Ada.Exceptions.Exception_Information(E));
        return OCI_Error;
   end Callback_In_Bind;
   
   function Callback_Out (Value : Value_Ptr;
                             bufpp   : access Buffer_ptr;
                             alenpp  : access Access_Size_T;
                             piecep  : access Ub1;
                             indpp   : access A_Sb2) return SWord
   is
      function Part return String is
      begin
     --    Text_IO.Put_Line (To_Ada(bufpp.all.all (0 .. alenpp.all.all-1), False));
         return To_Ada (bufpp.all.all (0 .. Value.alen - 1), False);
      end;
      use type A_Sb2;
   begin
--      text_io.put_line ("picep.all = " & Ub1'Image (piecep.all));
      case piecep.all is
      when OCI_ONE_PIECE | OCI_FIRST_PIECE =>
         pragma Debug ( Show (
         "out OCI_one_first_PIECE " & Ub1'Image (piecep.all)
            ));
            
         bufpp.all := Get_Buffer;

         Value.alen := Char_Buffer'Length;
         alenpp.all := Value.alen'Access;
         indpp.all := Value.Indicator'Access;
         Value.Prev_Indicator := Value.Indicator;
         indpp.all.all := Indicator_Unchanged;
         
         Value.Complete := False;
         Set_Last_Value (Value);
         Value.First_Out_Piece := True;

      when OCI_NEXT_PIECE =>
      
         pragma Debug ( Show ("out OCI_NEXT_PIECE " & size_t'Image (alenpp.all.all)));

         bufpp.all := Get_Buffer_Ptr;
         alenpp.all := Value.alen'Access;
         indpp.all := Value.Indicator'Access;
         
         if Value.First_Out_Piece then
            Value.Value := To_Unbounded_String (Part);
            Value.First_Out_Piece := False;
         else
            Append (Value.Value, Part);
         end if;

         if alenpp.all.all < Char_Buffer'Length then
            Value.Complete := True;
            Set_Last_Value (null);
         else
            pragma Assert (not Value.Complete, "somebody get buffer");
            null;
         end if;
         
      when OCI_LAST_PIECE =>
         pragma Debug ( Show ("out OCI_LAST_PIECE "));
         Append (Value.Value, Part);
         alenpp.all.all := 0;
      when others => 
--        text_io.put_line ("picep.all = " & Ub1'Image (piecep.all));
        return OCI_Error;
     end case;  
--     Text_IO.Put_Line (Boolean'Image (Value.Complete));
     return OCI_CONTINUE;
   exception
      when e: others =>
        Text_IO.Put_Line ("In callback " & Ada.Exceptions.Exception_Information(E));
        return OCI_Error;
   end Callback_Out;

   function Callback_Out_Bind (
                             octxp   : Value_Ptr;
                             bindp   : OCIBind;
                             iter    : Ub4;
                             index   : Ub4;
                             bufpp   : access Buffer_ptr;
                             alenpp  : access Access_Size_T;
                             piecep  : access Ub1;
                             indpp   : access A_Sb2;
                             rcodepp : access A_Ub2) return SWord
   is
   begin
      rcodepp.all := octxp.rcode'Access;
      return Callback_Out (octxp, bufpp, alenpp, piecep, indpp);
   end Callback_Out_Bind;
   
   function Callback_Define (
                                octxp   : Value_Ptr;
                                defnp   : OCIDefine;
                                iter    : Ub4;
                                bufpp   : access Buffer_Ptr;
                                alenpp  : access Access_Size_T;
                                piecep  : access Ub1;
                                indpp   : access A_Sb2;
                                rcodepp : A_A_Ub2) return SWord
   is
   begin
      pragma Debug (Show ("callback out index iter " & UB4'Image (iter)));
      rcodepp.all := octxp.rcode'Access;
      return Callback_Out (octxp, bufpp, alenpp, piecep, indpp);
   end Callback_Define;

   procedure Bind_Dynamic (Value : in out Var_Type'Class) is
      rc : SWord := OCIBindDynamic (
               bindp  => Value.Bind,
               errhp  => OCI.Thread.Error,
               ictxp  => Value'Address,
               icbfp  => Callback_In_Bind'Address,
               octxp  => Value'Address,
               ocbfp  => Callback_Out_Bind'Address);
   begin
      Check_Error(rc);
   end Bind_Dynamic;
  

   procedure Bind(Stmt : Statement; Value : in out Var_Type'Class; Name : String)
   is
       rc : sword := OCIBindByName (
                   stmtp => OCIStmt ( Handle (Stmt)),
                   bindpp => Value.Bind'Access,
                   errhp => OCI.Thread.Error,
                   placeholder => C.To_C(Name),
                   placeh_len => Name'Length,
                   valuep => System.Null_Address,
                   value_sz => String_Max_Size,
                   dty => Type_Id,
                   indp => Value.Indicator'Access,
                   mode => OCI_DATA_AT_EXEC);
   begin
     Check_Error (rc);
     Bind_Dynamic (Value);
   end Bind;

   procedure Bind (Stmt : Statement; Value : in out Var_Type'Class; Position : Positive)
   is
       rc : sword := OCIBindByPos (
                   stmtp => OCIStmt( Handle (Stmt)),
                   bindpp => Value.Bind'Access,
                   errhp => OCI.Thread.Error,
                   Position => ub4(Position),
                   valuep => System.Null_Address,
                   value_sz => String_Max_Size,
                   dty => Type_Id,
                   indp => Value.Indicator'Access,
                   mode => OCI_DATA_AT_EXEC);
   begin
     Check_Error(rc);
     Bind_Dynamic (Value);
   end Bind;

   procedure Define (Stmt : Statement;
      Value : in out Var_Type'Class;
      Position : Positive)
   is
      rc : sword := OCIDefineByPos (
                   stmtp => OCIStmt( Handle (Stmt)),
                   defnpp => Value.Define'Access,
                   errhp => OCI.Thread.Error,
                   position => ub4 (Position),
                   value => System.Null_Address,
                   value_sz => String_Max_Size,
                   dty => Type_Id,
                   indp => Value.Indicator'Access,
                   mode => OCI_DYNAMIC_FETCH);
   begin
     Check_Error(rc);
     rc := OCIDefineDynamic (
               defnp  => Value.Define,
               errhp  => OCI.Thread.Error,
               octxp  => Value'Address,
               ocbfp  => Callback_Define'Address);
     Check_Error(rc);
   end Define;

end OCI.Thick.Strings;