------------------------------------------------------------------------------
--  Ada95 Interface to Oracle RDBMS                                         --
--  Copyright (C) 2000-2001 Dmitriy Anisimkov.                              --
--  License text and authors contact information could be found in oci.ads  --
------------------------------------------------------------------------------
with
   Ada.Finalization,
   Ada.Task_Identification,
   Oci.Lib,
   Reference_Finalization;

--   uncomment it for debug printing
--   with Text_IO, System.Address_Image; use Text_IO,System;

package OCI.Thick is
   
   package RF renames Reference_Finalization;

   use Lib; 
   use Ada.Task_Identification;

   subtype OCIDate is Lib.OCIDate;

   subtype OCINumber is Lib.OCINumber;
   
   type Limited_Variable is tagged limited private;

   procedure Clear_Value (Var : in out Limited_Variable'Class);
   function Is_Null (Var : Limited_Variable'Class) return Boolean;
   
   INVALID_HANDLE : exception;
   LIB_ERROR      : exception;
   NULL_VALUE     : exception;
   NOT_ATTACHED   : exception; -- trying get not attached variable


   function Last_Error_Code return Integer;
   
   function Last_Error_Message (Record_No : Integer := 1) return String;
   
   -- Position of last error in the DML statement.
   function Last_Error_Offset return Natural;

-- OCIDate

   function To_String (From : OCIDate; Format : String) 
        return String;
   function SysDate return OCIDate;
        
   Init_OCIDate : constant OCIDate;

   type Handle_Reference is abstract new Rf.Controlled_Reference with private;
   
   procedure Check_Error (Code : Sword);

   Too_Late_For_Dedicated_Environment : exception;
  
   procedure Task_Dedicated_Environment;

   procedure Task_Done (T : Task_ID := Current_Task);
   -- Call it, when are you sure about using the task

   function Handle (ref : Handle_Reference'Class) return OCIHandle;
   pragma Inline (Handle);
   
private

   type Handle_Reference is abstract 
        new Rf.Controlled_Reference with record
      Handle : OCIHandle := Empty_Handle; 
   end record;

   package AF renames Ada.Finalization;

   type Limited_Variable is new AF.Limited_Controlled with record
      Indicator : aliased Sb2 := Null_Indicator;
      Bind     : aliased OCIBind := OCIBind (Empty_Handle);
      Define  : aliased OCIDefine := OCIDefine (Empty_Handle);
   end record;

   procedure Free (H : OCIHandle; HType : Ub4);
   function Alloc_Handle (Parent : OCIEnv; Htype : Ub4) return OCIHandle;
   function Alloc_Descriptor (Parent : OCIEnv; Htype : Ub4) return OCIHandle;
   
   function Get_Attr (Param : OCIHandle; HType : ub4; Attr : Ub4) return Integer;
   function Get_Attr (Param : OCIHandle; HType : ub4; Attr : Ub4) return String;

--   procedure Set_Attr (H : OCIHandle; HType : ub4; Attr : Ub4; Value : Integer);
   procedure Set_Attr (H : OCIHandle; HType : ub4; Value : String; Attr : Ub4);
   
   procedure Set_Attr (H : OCIHandle; HType : ub4; Value : OCIHandle; Attr : Ub4);

   Init_OCIDate : constant OCIDate := (
      OCIDateYYYY => 0, -- gregorian year; range is -4712 <= year <= 9999
      OCIDateMM   => 1, -- month; range is 1 <= month < 12
      OCIDateDD   => 1, -- day; range is 1 <= day <= 31
      OCIDateTime => (
         OCITimeHH => 0, -- hours; range is 0 <= hours <=23
         OCITimeMI => 0, -- minutes; range is 0 <= minutes <= 59
         OCITimeSS => 0) -- seconds; range is 0 <= seconds <= 59
     );    

end OCI.Thick;